import { MPC } from 'mpc-js';
import OBSWebSocket from 'obs-websocket-js'
import { MidiController } from './midiController.js'
import { paBus } from './pulseaudioBus.js'
import { spawn } from 'child_process'

// USER SETTINGS
const midiDeviceSubstring = 'LPD8'

const SCENE_CC_MAP = new Map([
  [11, 'AFK'],
  [12, 'Screen'],
])

const AUDIO_CC_MAP = new Map([
  [15, 'rubix12'],
  [16, 'rubix12game'],
  [18, 'rubix12music'],
  [19, 'rubix12voip'],
])
const AUDIO_VOL_MAP = new Map([
  [1, 'rubix12'],
  [2, 'rubix12game'],
  [3, 'rubix12music'],
  [4, 'rubix12voip'],
])
// end USER SETTINGS

// Derived values
const AUDIO_CC_MAP_INV = new Map([...AUDIO_CC_MAP].map(([k, v]) => [v, k]))
const AUDIO_VOL_MAP_INV = new Map([...AUDIO_VOL_MAP].map(([k, v]) => [v, k]))
const SCENE_CC_MAP_INV = new Map([...SCENE_CC_MAP].map(([k, v]) => [v, k]))

// State
const audioNameIdMap = new Map()
const audioIdNameMap = new Map()
let focusedClients = null

// Libraries initialization
const obs = new OBSWebSocket();
const mpc = new MPC()
const pa = paBus(AUDIO_CC_MAP.values())
const midiController = MidiController(midiDeviceSubstring)

pa.on('deviceUpdate', ({ name, index }) => {
  audioNameIdMap.set(name, index)
  audioIdNameMap.set(index, name)
})

pa.on('error', () => {
  audioIdNameMap.clear()
  audioNameIdMap.clear()
  focusedClients = null
})

mpc.on('ready', async () => {
  console.log('MPC Ready')
})

// Looping shit
let timeout
const syncScenes = (currentScene) => {
  SCENE_CC_MAP.forEach((scene, cc) => {
    if (timeout !== undefined) clearTimeout(timeout)
    const enable = scene === currentScene ? 1 : 0
    midiController.setControl(0, cc, enable)
    // Handle the case when we click already enabled scene
    if (enable) {
      timeout = setTimeout(() => {
        midiController.setControl(0, cc, enable)
      }, 900)
    }
  })
}
let timeout2
const syncDevices = (currentDevices) => {
  AUDIO_CC_MAP.forEach((device, cc) => {
    if (timeout2 !== undefined) clearTimeout(timeout2)
    const enable = new Set((currentDevices || []).map(({ sink }) => audioIdNameMap.get(sink))).has(device) ? 1 : 0
    midiController.setControl(0, cc, enable)
    // Handle the case when we click already enabled scene
    if (enable) {
      timeout2 = setTimeout(() => {
        midiController.setControl(0, cc, enable)
      }, 900)
    }
  })
}

// await Promise.all([
//   obs
//     .connect({
//       address: 'localhost:4444',
//       password: 'cz1jJgeBaeolQCHhgCZM'
//     })
//     .then(() => {
//       console.info(`Success! OBS connected & authenticated.`);
//       return obs.send('GetSceneList');
//     })
//     .then((a) => {
//       syncScenes(a.currentScene)
//     })
//     .catch((e) => console.error(e)),
//   mpc.connectTCP('rpi.lan', 6600),
//   pa.promise
// ]);

console.info('All services connected')

SCENE_CC_MAP_INV.forEach((controlId, scene) => {
  midiController.on(`0:ControlChange:${controlId}`, ({ value }) => {
    value !== 0 && obs.send('SetCurrentScene', { 'scene-name': scene })
  })
})

AUDIO_CC_MAP_INV.forEach((controlId, device) => {
  midiController.on(`0:ControlChange:${controlId}`, ({ value }) => {
    if (focusedClients !== null && value !== 0) {
      focusedClients.forEach((focusedClient) => {
        pa.moveSinkInput(focusedClient.index, device)
      })
    }
  })
})

AUDIO_VOL_MAP_INV.forEach((controlId, device) => {
  midiController.on(`0:ControlChange:${controlId}`, ({ value }) => {
    pa.setSinkVolume(value, device)
  })
})

const getFocusPid = async () => {
  const xdo = spawn('xdotool', ['getwindowfocus', 'getwindowpid'])
  let pid
  for await (const data of xdo.stdout) {
    pid = Number(data)
  };
  return pid
}

const getFocusInput = async () => {
  const pid = await getFocusPid()
  const list = await pa.getSinkInputList()
  const ret = list.filter(i => i.properties.application).filter(i => i.properties.application.process).filter(i => Number(i.properties.application.process.id) === pid)
  return ret
}

let timeout3
const loopCheck = async () => {
    if (timeout3 !== undefined) clearTimeout(timeout3)
  const focusedInputs = await getFocusInput()
  const fc = focusedInputs.map(({index, sink}) => ({ index: Number(index), sink }))
  if (fc.length > 0) {
    focusedClients = fc
    syncDevices(fc)
  } else {
    focusedClients = null
    syncDevices(null)
  }
  timeout3 = setTimeout(loopCheck, 1000)
}
timeout3 = setTimeout(loopCheck, 1000)

obs.on('SwitchScenes', ({ sceneName }) => {
  if (SCENE_CC_MAP_INV.has(sceneName)) {
    syncScenes(sceneName)
  }
})

obs.on('Exiting', err => {
    console.error('Exiting', err);
});

obs.on('error', err => {
    console.error('socket error:', err);
});

midiController.on('0:ControlChange:5', ({ value }) => {
  mpc.playbackOptions.setVolume(Math.floor(value * 100))
  mpc.playback.pause(value === 0)
  obs.send('SetSceneItemRender', { 'scene-name': 'AFK', source: 'Muzak', render: value !== 0 })
})

console.info('All systems go.')
