import midi from '@julusian/midi'
import EventEmitter from 'events'

export const MidiController = (midiDeviceSubstring) => {
  // All nibbles used for voice/control midi messages
  const NOTE_OFF = 0x8
  const NOTE_ON = 0x9
  const POLY_AFTERTOUCH = 0xA
  const CONT_CHANGE = 0xB
  const PROG_CHANGE = 0xC
  const AFTERTOUCH = 0xD
  const PITCHBEND = 0xE
  // input.ignoreTypes(true, false, true)
  const midiController = new EventEmitter()

  const input = new midi.Input()
  const output = new midi.Output()
  midiController.connected = false;

  midiController.findDevices = () => {
    const portTotal = input.getPortCount()
    // Order: (Sysex, Timing, Active Sensing)
    input.ignoreTypes(false, false, false)

    let midiDevicePort = -1
    for (let i = 0; i < portTotal; i++) {
      const name = input.getPortName(i)
      if (name.includes(midiDeviceSubstring)) {
        midiDevicePort = i
        break
      }
    }
    return midiDevicePort
  }

  midiController.init = (midiDevicePort) => {
    input.openPort(midiDevicePort)
    output.openPort(midiDevicePort)
    midiController.connected = true

    // Attempt to reset (Doesn't work LPD8)
    output.sendMessage([0xFF])
  }

  input.on('message', (deltaTime, message) => {
    // https://www.cs.cf.ac.uk/Dave/Multimedia/node158.html has some helpful
    // information interpreting the messages.
    const [status, data1, data2] = message;
    const func = (status & 0xF0) >> 4
    const chan = status & 0x0F
    if (chan === 0) {
      switch (func) {
        case CONT_CHANGE:
          const controlId = data1
          const value = data2 / 127
          midiController.emit(`${chan}:ControlChange:${controlId}`, { value, deltaTime, controlId })
          break;
        case NOTE_OFF:
        case NOTE_ON:
          const key = data1
          const force = data2 / 127
          const isRelease = func === NOTE_OFF
          midiController.emit(`${chan}:Note:${key}`, { force, isRelease, deltaTime, key })
          break
        case PROG_CHANGE:
          const program = data1
          midiController.emit(`${chan}:ProgramChange:${program}`, { deltaTime, program })
          break
      }
    }
  })

  midiController.setControl = (channel, controllerId, value) => {
    output.sendMessage([(CONT_CHANGE << 4) + channel, controllerId, value])
  }

  midiController.setNote = (channel, pressed, key, force) => {
    output.sendMessage([((pressed ? NOTE_ON : NOTE_OFF) << 4) + channel, key, force])
  }

  const searchForDevices = () => {
    const port = midiController.findDevices()
    if (port > 0) {
      if (!midiController.connected) {
        midiController.init(port)
      }
    } else {
      input.closePort()
      output.closePort()
      midiController.connected = false
    }
    setTimeout(searchForDevices, 1000)
  }

  searchForDevices()

  return midiController
}
